/* jshint unused: true, moz: true*/

Components.utils.import("resource://gre/modules/AddonManager.jsm");
Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
const ci = Components.interfaces;

var restore = null; // for pref (restore hilite after closing AoM)
var showUI = null; // for pref (whether to create/display ac menu)
var UIwidth = null; // for pref (ac menu size)
var showVersionNumber = null; // for pref (whether to show version number in ac menu)
var lastUpdatedStateGrouping = null; // for pref (whether to group lastUpdated sort by enabled/disabled)
var spacer = "   ";
var extraRows = null;
var aomhiliteMap = null;

var addonsArray = [];
var addonCatArray = [];
var categoryList = null;

var savedAddonName = null;
var displayDropdown = null; // used for Seamonkey & Tbird workaround

function logthis(entry) { Services.console.logStringMessage(entry); } 

function startup(data, reason) {
  Services.obs.addObserver(ChromeDocObserver, "chrome-document-global-created", false);
  var defPrefs = Services.prefs.getDefaultBranch("extensions.addonsmgrhilte@cfl.");
  defPrefs.setBoolPref("restore", true);
  defPrefs.setBoolPref("extraRows", false);
  defPrefs.setBoolPref("showUI", true);
  defPrefs.setIntPref("listboxSize", 18);
  defPrefs.setBoolPref("showVersion", false);
  defPrefs.setBoolPref("lastUpdatedStateGrouping", false);  
  defPrefs.setCharPref("aomhiliteSavedSelections", "");
  
  // set up main stylesheet (includes styles for sort buttons & acAddonList)
  addStylesheet("chrome://aomhilite/content/aomhilite.css");
  
  if (reason != APP_STARTUP) {
    //logthis("finding and reloading Add-ons Managers");   
    var mgrs = findAddonsMgrs();
    var i;
    for (i = 0; i < mgrs.length; i++) {
      mgrs[i].location.reload();
    }
  }  
}

function shutdown(data, reason) {
  Services.obs.removeObserver(ChromeDocObserver, "chrome-document-global-created");

  // call flush due to the bundle service cache (for aomhilite.properties string bundle)
  Components.classes["@mozilla.org/intl/stringbundle;1"]
                    .getService(ci.nsIStringBundleService)
                    .flushBundles();

  // unregister our custom autocomplete component  and unload its module
  if (typeof unregComponent != 'undefined') {
    unregComponent();
    Components.utils.unload("chrome://aomhilite/content/jsm/autocomplete.jsm");
  }
  
  // remove stylesheets
  removeStylesheet("chrome://aomhilite/content/aomhilite.css");
  removeStylesheet("chrome://aomhilite/content/aomhiliteLG.css");
  removeStylesheet("chrome://aomhilite/content/aomhiliteOSX.css");

  aomhilitePrefObserver.unregister();
  aomhilitePrefObserver = null;

  // stuff gets cleaned up on Fx restart anyway ???
  //if (reason == APP_SHUTDOWN) { return; }
  // removes any UI from any open Add-ons Manager tabs, reverts to default sorting, triggers cleanup
  //logthis("finding and reloading Add-ons Managers");   
  var mgrs = findAddonsMgrs();
  var i;
  for (i = 0; i < mgrs.length; i++) {
    var win = mgrs[i];
    // hide them to prevent flash of unstyled buttons
    var sortButtons = win.document.getElementById("aomhilite-sorters");
    sortButtons.setAttribute("hidden", true);
    win.location.reload(true);
  }
}

function install(data, reason) {}

function uninstall(data, reason) {}

function ChromeDocObserver(aSubject, aTopic, aData) {
  try {
    if (aTopic != "chrome-document-global-created") { return; }
    var win = aSubject;
    var uri = win.document.documentURI;
    if (uri == "about:addons") {    
      win.addEventListener("load", AddonsMgrOnload, false);
    }
  } catch (e) {
    logthis("Exception in ChromeDocObserver");
  }
}

function AddonsMgrOnload() {
  // removing this load listener is req'd to prevent mem leak
  // listener no longer needed since if the mgr is reloaded via F5,
  // unload is called, then ChromeDocObserver re-adds the listener
  this.removeEventListener("load", AddonsMgrOnload, false);
  this.addEventListener("unload", AddonsMgrOnUnload, false); 
  var AMdoc = this.document; // document obj of Addons Mgr window
  var addonList = AMdoc.getElementById("addon-list");
  var prefs = Components.classes["@mozilla.org/preferences-service;1"]
              .getService(ci.nsIPrefService);
  prefs = prefs.getBranch("extensions.addonsmgrhilte@cfl.");
  restore = prefs.getBoolPref("restore");
  showUI = prefs.getBoolPref("showUI");
  extraRows = prefs.getBoolPref("extraRows");
  showVersionNumber = prefs.getBoolPref("showVersion");
  lastUpdatedStateGrouping = prefs.getBoolPref("lastUpdatedStateGrouping");    
  aomhiliteMap = prefs.getCharPref("aomhiliteSavedSelections");  
  UIwidth = prefs.getIntPref("listboxSize");
  UIwidth = (UIwidth > 50) ? 50 : UIwidth;
  UIwidth = (UIwidth < 10) ? 10 : UIwidth;
  aomhilitePrefObserver.register();

  if (!restore) { 
    // re-initialze list of saved hilites if not 'always restore'
    var map = {};
    map = JSON.parse(aomhiliteMap); 
    var sortby = map.sortby;
    map = {};               // delete saved hilites
    map.sortby = sortby; // but save 'sortby'
    addonList.setAttribute('aomhilite', JSON.stringify(map));
    //logthis(JSON.stringify(map));
    //    addonList.setAttribute('aomhilite', "");  // possible future undocumented pref?
  } else {
    addonList.setAttribute('aomhilite', aomhiliteMap);
  }    

  this.addEventListener("ViewChanged", onViewChange, false);

  var cat = AMdoc.getElementById("categories");
  if (!cat || !addonList) { return; }
  addonList.addEventListener("select", onSelect, false);
  initAddonList(AMdoc);
  if (showUI) { initUI(AMdoc); }
}

function initAddonList (AMdoc) {
  // check if we're in list view of a supported category; sort and hilite
  var selCat = getSelectedCat(AMdoc);
  
  // init our sorting buttons
  var sortButtons = AMdoc.createElement("hbox");
  sortButtons.setAttribute("id", "aomhilite-sorters");
  sortButtons.setAttribute("class", "view-header sort-controls");
  var listView = AMdoc.getElementById("list-view");
  var addonList = AMdoc.getElementById("addon-list");
  listView.insertBefore(sortButtons, addonList);
  // repurpose price button
  var button = AMdoc.getAnonymousElementByAttribute(sortButtons, "anonid", "price-btn");
  button.setAttribute("anonid", "state-btn");
  button.setAttribute("hidden", "false");
  button.setAttribute("label", "Enabled");
  button.setAttribute("tooltiptext", "Sort by Enabled/Disabled");
  button.setAttribute("oncommand", "this.parentNode._handleChange('state');");
  sortButtons.addEventListener("command", sortButtonHandler, true);
  
  if (getCurrentViewType(AMdoc) == "list-view" ) {  
    // add-ons known to have own sort buttons
    if (selCat == "userstyle"       ||            // Stylish
        selCat == "custombuttons"   ||            // Custom Buttons
        selCat == "userscript"      ||            // Scriptish (fork of Greasemonkey)
        selCat == "greasemonkey-user-script") {   // Greasemonkey
      sortButtons.setAttribute("hidden", true);
    } 

    // this is not saved automatically (we are after all dynamically creating it)
    // we need to read/save this on unload; restore it on load
    // and remove it from the DOM on unload too
    var savedSelections = addonList.getAttribute('aomhilite');    
    var map = {};
    if (savedSelections) { map = JSON.parse(savedSelections); } 
    var sortby = map.sortby;
    if (!sortby) { sortby = "state"; }  // 'state' is the default
    var sorters = AMdoc.getElementById("aomhilite-sorters");
    // sets sortby attr, checkState, thus enabling corresponding button to look activated    
    sorters.setSort(sortby);  
    // setSort handles (un)setting the checked state from the other buttons 
    if (sortby == "state") { 
      button.setAttribute("checkState", 1); 
      button.checked = true;
    }
    
    //logthis("sortby: " + sortby);    
    if (sortby != "state") {  // "state" is the default sort
      // can we just 'click' the desired button instead?
      resortAddons(AMdoc, sortby); 
    }
    restoreHilite(AMdoc);
  }
}

var addonListener = {
  // re-get the list of add-ons when one has been installed/uninstalled
  onInstalled : function(addon) {
    refreshACaddonList();  
  },

  /*
   * due to the 'undo' option remaining available until changing categories or aom closed/reopened, 
   * this won't get called until then, because that's when the actual uninstall occurs 
   * (and when we should remove item from  dropdown list)
   */
  onUninstalled : function(addon) {
    // obviously unnecessary, if it's this addon that's being uninstalled
    if (addon.name == "Addons Manager Hilite") { return; }
    refreshACaddonList();
  }
};

function refreshACaddonList() {
  addonsArray = [];
  categoryList = null;
  AddonManager.getAllAddons(function(addons) {
      addonListCallback(addons, null);
    });    
}

function initUI(AMdoc) {
  // load and register our custom autocomplete component
  if (typeof regComponent == 'undefined') {
    Components.utils.import("chrome://aomhilite/content/jsm/autocomplete.jsm");
    regComponent();
  }
  
  var acListboxStyle = "margin-left: 1em; margin-right: 1em; width: " + UIwidth + "em;";
  var acAddonList = AMdoc.createElement("textbox");
  acAddonList.setAttribute("id", "aomhilite_ac_menu");
  acAddonList.setAttribute("style", acListboxStyle);
  acAddonList.setAttribute("type", "autocomplete");
  // acAddonList.setAttribute("size", 24); // seems to have no effect; set via css width instead
  acAddonList.setAttribute("autocompletesearch", "addonList_autocomplete");
  acAddonList.setAttribute("enablehistory", true);
  acAddonList.setAttribute("forcecomplete", true); // ensures selection is valid
  // only applies to typing in box; see aomhilite.xml
  acAddonList.setAttribute("maxrows", 10); // default is about 6
  var notifyListbox = "var evt = document.createEvent('Events'); " 
                      + "evt.initEvent('textentered', true, true); " 
                      + "this.dispatchEvent(evt);";
  acAddonList.setAttribute("ontextentered", notifyListbox);

  var strbundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
      .getService(ci.nsIStringBundleService)
      .createBundle("chrome://aomhilite/locale/aomhilite.properties");
  var selectaddon = strbundle.GetStringFromName("placeholderString");
  acAddonList.setAttribute("placeholder", selectaddon);

  /* set up OSX stylesheet */
  var osString = Components.classes["@mozilla.org/xre/app-info;1"].getService(ci.nsIXULRuntime).OS;
  if (osString == "Darwin") {
    addStylesheet("chrome://aomhilite/content/aomhiliteOSX.css");
  }

  var updateBtn = AMdoc.getElementById("header-utils-btn");
  var head = AMdoc.getElementById("header");
  head.insertBefore(acAddonList, updateBtn);
 
  // prevents "Extension cannot be installed because Nightly cannot modify the needed file"
  // error when updating this ext (bug 719180)
  // although 71980 is fixed in fx 19+, using (tested on fx 21.0a1):
  // acAddonList.style.MozBinding = "url(chrome://aomhilite/content/aomhilite.xml#ac_menu_aomhilite)";
  // leaves the aomhilite.xml file in memory even when AoM is closed and aomhilite is disabled!

  var chromeReg = Components.classes["@mozilla.org/chrome/chrome-registry;1"]
                  .getService(ci.nsIChromeRegistry);
  var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                  .getService(ci.nsIIOService);
  var bindingUri;
  if (extraRows) {
    bindingUri = ioService.newURI("chrome://aomhilite/content/aomhiliteLG.xml#ac_menu_aomhilite", null, null);
  } else {
    bindingUri = ioService.newURI("chrome://aomhilite/content/aomhilite.xml#ac_menu_aomhilite", null, null);
  }
  var convertedBindingUri = chromeReg.convertChromeURL(bindingUri);
  acAddonList.style.MozBinding = 'url("' + convertedBindingUri.spec + '")';

  // workaround - Seamonkey (bug 708520) not opening popup by default (Tb fixed as of Tb31 - Bug 360648)
  const SEAMONKEY_ID = "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}";
  var appInfo = Components.classes["@mozilla.org/xre/app-info;1"].getService(ci.nsIXULAppInfo);
  if (appInfo.ID == SEAMONKEY_ID) {
    displayDropdown = function(e) {
      var id = e.originalTarget.getAttribute("anonid");
      if (id != "historydropmarker") { return; }
      var acAddonList = e.target.ownerDocument.getElementById("aomhilite_ac_menu");
      acAddonList.showHistoryPopup();
    };
      acAddonList.addEventListener("click", displayDropdown, false);
  }
  acAddonList.addEventListener("textentered", onACaddonListSelection, false);

  AddonManager.addAddonListener(addonListener);
  // (do here since addonListCallback will require data initialized above)
  AddonManager.getAllAddons(function(addons) {
      addonListCallback(addons, AMdoc);
    });
}

function addStylesheet(sheet) {
  var sss = Components.classes["@mozilla.org/content/style-sheet-service;1"]
            .getService(ci.nsIStyleSheetService);
  var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(ci.nsIIOService);
  var uri = ios.newURI(sheet, null, null);
  if (!sss.sheetRegistered(uri, sss.USER_SHEET)) {
    sss.loadAndRegisterSheet(uri, sss.USER_SHEET);
  }
}

function removeStylesheet(sheet) {
  var sss = Components.classes["@mozilla.org/content/style-sheet-service;1"]
            .getService(ci.nsIStyleSheetService);
  var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(ci.nsIIOService);
  var uri = ios.newURI(sheet, null, null);
  if (sss.sheetRegistered(uri, sss.USER_SHEET)) {
    sss.unregisterSheet(uri, sss.USER_SHEET);
  }
}

function findAddonsMgrs() {
  // find all open mgrs
  var mgrs = [];
  const Tb_ID = "{3550f703-e582-4d05-9a08-453d09bdfdc6}";
  var appInfo = Components.classes["@mozilla.org/xre/app-info;1"].getService(ci.nsIXULAppInfo);
  var windowtype = (appInfo.ID == Tb_ID) ? null : "navigator:browser"; 
  var windows = Services.wm.getEnumerator(windowtype);
  var window;
  while (windows.hasMoreElements()) {
    window = windows.getNext().QueryInterface(ci.nsIDOMWindow);
    if (appInfo.ID == Tb_ID) {  // Thunderbird 
      var tabmail = window.document.getElementById("tabmail");
      for (var i = 0; i < tabmail.tabInfo.length; i++) {
        var tab = tabmail.tabInfo[i];
        if (tab.mode.type == "contentTab" && 
            tab.browser.contentDocument.location.href == "about:addons") {
          mgrs.push(tab.browser.contentWindow);
        }
      }
      return mgrs;
    }
    window.gBrowser.browsers.forEach(findMgrTab); // Fx & Seamonkey
  }
  function findMgrTab(browser) {
    if (browser.currentURI.spec == "about:addons") {
      mgrs.push(browser.contentWindow);
    }
  }
  return mgrs;
}

function updateAllListboxes() {
  var mgrs = findAddonsMgrs();
  var i;
  for (i = 0; i < mgrs.length; i++) {
    var AMdoc = mgrs[i].document;
    var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
    // clear this in case we are adding/removing version numbers
    if (acAddonList) { acAddonList.value = ""; }    
    var selCat = getSelectedCat(AMdoc);
    addonCatArray = addonsArray[categoryList.indexOf(selCat)];
    acAddonList.setAttribute("autocompletesearchparam", JSON.stringify(addonCatArray));
  }
}

function restoreHilite(AMdoc) {
  // purpose: retrieve saved selection and restore it
  // content.document does not get document we want if AoM was opened in background via session restore
  var selAddon;
  var addonList = AMdoc.getElementById('addon-list');
  var savedSelections = addonList.getAttribute('aomhilite');
  var map = {};
  if (savedSelections) { map = JSON.parse(savedSelections); }
  var selCat = getSelectedCat(AMdoc);
  var lastSel = map[selCat];
  if (lastSel) {
    selAddon = addonList.getElementsByAttribute('value', lastSel);
  }
  if (!lastSel || !selAddon || !selAddon[0]) {
    // no lastSel will occur on newly opened AoM if (always) restore pref is false
    // otherwise typically occurs when selected category has no installed items or has never had a saved selection
    var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
    if (acAddonList) {
      // onSelect not triggered (see below); need to clear this now
      acAddonList.value = "";
    }
    return;
  }
  addonList.selectItem(selAddon[0]); // will trigger onSelect
  selAddon[0].scrollIntoView();
  addonList.focus();
}

function saveUpdatedSel(addonList, selCat, newSelId) {
  var savedSelections = addonList.getAttribute('aomhilite');
  var map = {};
  if (savedSelections) { map = JSON.parse(savedSelections); } 
  selCat = selCat.slice(selCat.indexOf("-") + 1); // remove the 'category-' portion
  map[selCat] = newSelId;
  var updatedSel = JSON.stringify(map);
  addonList.setAttribute('aomhilite', updatedSel);
}

function onSelect(e) { // called once per add-on upon loading mgr, but (this.currentIndex = = -1) at that point
  // called when clicking an entry in addonList, selecting from autocomplete list, and from restoreHilite funct; also when changing categories
  if (e.eventPhase != e.AT_TARGET || !this.selectedItem) { return; }
  if (this.currentIndex < 0) { return; }  // there is no current selected item in this category
  var currentSel = this.selectedItem.mAddon.id;
  var AMdoc = this.ownerDocument;
  var selCat = getSelectedCat(AMdoc);
  if (!currentSel) { return; }
  saveUpdatedSel(this, selCat, currentSel);
  var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
  if (!acAddonList || !acAddonList.value) { return; }
  if (acAddonList.value.indexOf(this.selectedItem.mAddon.name) < 0) {
    acAddonList.value = ""; // clear ac list input field, rather than show name of a now unselected item
  }
}

function onACaddonListSelection() {
  // called when selection made from ac dropdown list or pressing ENTER in input field
  var AMdoc = this.ownerDocument;
  var cat = AMdoc.getElementById('categories');
  var selCat = cat.selectedItem.id;
  var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
  // setting selectionStart & selectionEnd to same value sets the cursor position to that position
  // set to beginning so text looks cropped on right instead of on left
  acAddonList.inputField.selectionStart = 0;
  acAddonList.inputField.selectionEnd = 0;
  var sel = acAddonList.value;
  
  if (showVersionNumber) {
    let
    nameEnd = sel.lastIndexOf(spacer);
    if (nameEnd > 0) { // remove spacer and version #, if found (should be there, but...)
      sel = sel.slice(0, nameEnd);
    }
  }
  var addonList = AMdoc.getElementById('addon-list');
  var selAddon = addonList.getElementsByAttribute('name', sel);
  if (selAddon && selAddon[0]) {
    var newSelId = selAddon[0].mAddon.id;
    if (!newSelId) { return; }
    saveUpdatedSel(addonList, selCat, newSelId);
    if (getCurrentViewType(AMdoc) == "list-view") {
      restoreHilite(AMdoc);
    } else {
      selAddon[0].showInDetailView();
    }
  } else {
    /*
     * Workaround issue that occurs when using nav arrow takes us back into detail view of a non-current category,
     * resulting in 'addon-list' element that is no longer populated with the list of add-ons, thus there are no
     * elements with the name we are looking for
     */
    savedAddonName = sel;
    // repopulates the list, finish workaround in the resulting onViewChange
    cat.selectedItem.click();
  }
}

function onViewChange() {
  var AMdoc = this.document;
  var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
  var viewType = getCurrentViewType(AMdoc);
  if (viewType == "unsupported") {
    if(acAddonList) {
      acAddonList.setAttribute('hidden', true);
    } 
    return; 
  }
  var cat = AMdoc.getElementById('categories');
  var selCat = cat.selectedItem.id;
  var sorters = AMdoc.getElementById("aomhilite-sorters");
  var dontSort = false;  
  if (selCat == "category-userstyle" || 
      selCat == "category-custombuttons" || 
      selCat == "category-greasemonkey-user-script") {
    dontSort = true;
    sorters.setAttribute("hidden", true);
  } else {
    sorters.removeAttribute("hidden");
  }
  if (viewType == "list-view") {
    if (savedAddonName) {
      // finish workaround from onACaddonListSelection - display add-on in detail view
      var addonList = AMdoc.getElementById('addon-list');
      // prevents annoying flash of list view before detail view appears
      addonList.setAttribute('hidden', true);
      var selAddon = addonList.getElementsByAttribute('name', savedAddonName);
      if (selAddon && selAddon[0]) {
        selAddon[0].showInDetailView();
        var newSelId = selAddon[0].mAddon.id;
        saveUpdatedSel(addonList, selCat, newSelId);
      }
      addonList.removeAttribute('hidden');
      savedAddonName = null;
      // workaround finished, we're now in detail view, which retriggers viewchange
      // (if we unexpectedly fail to find add-on for detail view, just continue in list view)
      if (selAddon && selAddon[0]) { return; }
    }

    if (!dontSort) {
      var sortby = sorters.getAttribute("sortby"); 
      if (sortby != "state") {  // 'state' is the default sort
        resortAddons(AMdoc, sortby);
      }
    }
    restoreHilite(AMdoc);
  } else {  // detail-view
    // if we are in this ext's detail view - test if still needed w/ options.xul code for this
    if (AMdoc.getElementById("detail-name").textContent == "Addons Manager Hilite") {
      // set input field of numeric setting to sensible size & disable if not applicable
      var setting = AMdoc.getElementById("aomhilite_size");
      var inputField = AMdoc.getAnonymousElementByAttribute(setting, "anonid", "input");
      inputField.setAttribute("style", "max-width: 3em;");
    }
  }

  // if our menu has not been created yet or the UI is pref'ed off, nothing else to do
  if (!acAddonList || !categoryList) { return; }
  
  selCat = selCat.slice(selCat.indexOf("-") + 1); // remove the 'category-' portion
  addonCatArray = addonsArray[categoryList.indexOf(selCat)];
  acAddonList.setAttribute("autocompletesearchparam", JSON.stringify(addonCatArray));
  // display ac menu only in categories w/ add-ons of that type
  if (categoryList.indexOf(selCat) != -1) {
    acAddonList.removeAttribute('hidden'); 
  } else {
      acAddonList.setAttribute('hidden', true);
  }
}

function sortButtonHandler(e) {
  var target = e.originalTarget;
  target.setAttribute("checkState", 1);
  target.checked = true;
  var anonid = target.getAttribute("anonid");
  var AMdoc = target.ownerDocument;
  var sortby = anonid.substring(0, anonid.indexOf("-"));
  if (sortby == "date") { sortby = "updateDate"; }
  resortAddons(AMdoc, sortby);
  e.preventDefault(); //prevents state button being underlined as selected, but otherwise causes freeze
  // something is altering checkState; restore it to being selected (look into what is causing alteration)
  if (sortby == "state") { 
    AMdoc.defaultView.setTimeout(function() {
      target.setAttribute("checkState", 1); 
      target.checked = true;
    }, 0);
  }  
  //e.stopPropagation();  // prevents sortby state effecting proper sort
}

function resortAddons(AMdoc, sortby) {
  var win = AMdoc.defaultView;
  var addonList = AMdoc.getElementById('addon-list');  
  switch (sortby) {
    case "state":
      win.gViewController.currentViewId = "";
      var selCat = getSelectedCat(AMdoc);
      var updatedView = "addons://list/" + selCat;  // e.g. addons://list/extension
      win.gViewController.replaceView(updatedView);
      break;
    case "name":
      var sortbyArray = [ "name" ];
      AMdoc.defaultView.sortList(addonList, sortbyArray, true); // true = ascending
      break;      
    case "updateDate": // most recently updated
      var list = addonList.children;
      list.sort(function(addonItem1, addonItem2) {
        var addon1 = addonItem1.mAddon;
        var addon2 = addonItem2.mAddon;
        var addon1Time = Date.parse(addon1.updateDate);
        var addon2Time = Date.parse(addon2.updateDate);
        if (lastUpdatedStateGrouping && addon1.isActive && !addon2.isActive) { return -1; }
        if (lastUpdatedStateGrouping && addon2.isActive && !addon1.isActive) { return 1; }        
        if (addon1Time > addon2Time) { return -1; }
        if (addon1Time < addon2Time) { return 1;  }
        var addon1Name = addon1.name.toLowerCase();
        var addon2Name = addon2.name.toLowerCase();
        return addon1Name.localeCompare(addon2Name);
      });
      var i;
      for (i = 0; i < list.length; i++) {
        addonList.appendChild(list[i]);
      }
      // unfortunately this throws an error
      //var sortby = [ "updateDate" ];
      //win.sortList(addonList, sortby, true); // true = ascending
      break;
  }    
}

function getCurrentViewType(AMdoc) {
  var win = AMdoc.defaultView;
  var controller = win.gViewController;
  // displayedView (only available in Fx48+) yields e.g. list-view
  // in parseViewId: .type yields e.g. list
  //                       .param yields e.g. extension
  var viewType = (controller.displayedView) ? controller.displayedView.id :
                                              controller.parseViewId(controller.currentViewId).type;
  if (viewType == "list" || viewType == "list-view") { 
    return "list-view"; 
  } else if (viewType == "detail" || viewType == "detail-view") {
    return "detail-view"; 
  } else { 
    return "unsupported"; 
  }
}

function getSelectedCat(AMdoc) {
  var cat = AMdoc.getElementById("categories");
  var selCat = cat.selectedItem.id;
  return selCat.slice(selCat.indexOf("-") + 1); // remove the 'category-' portion  
}

function addonListCallback(addons, AMdoc) {
  addons.sort(function(addonItem1, addonItem2) {
    var addon1Name = addonItem1.name.toLowerCase();
    var addon2Name = addonItem2.name.toLowerCase();
    if (addonItem1.type == addonItem2.type){
      return addon1Name.localeCompare(addon2Name);
      } else {
      return addonItem1.type.localeCompare(addonItem2.type);
    }    
  });

  var currentType = addons[0].type;
  var addonTypeArray = [];
  categoryList = [currentType];
  var i;  
  for (i = 0; i < addons.length; i++) {
    //logthis(addons[i].name + ": " + addons[i].type);
    if (addons[i].hidden) { 
      //logthis("Addons Manager Hilite is ignoring hidden add-on: " + addons[i].name); 
      continue;
    }
    var newType = addons[i].type;
    if (newType == currentType) {
      if (showVersionNumber && addons[i].version) {
        addonTypeArray.push(addons[i].name + spacer + addons[i].version);
      } else {
        addonTypeArray.push(addons[i].name);
      }
    } else {
      // use slice to get a copy of the array, instead of a ref to it
      addonsArray.push(addonTypeArray.slice()); 
      addonTypeArray.length = 0;
      if (showVersionNumber && addons[i].version) {
        addonTypeArray.push(addons[i].name + spacer + addons[i].version);
      } else {
        addonTypeArray.push(addons[i].name);
      }
      categoryList.push(newType);
      currentType = newType;  
    }
  }
  addonsArray.push(addonTypeArray.slice());   
  if (!AMdoc) { // null when called from refreshACaddonList
    updateAllListboxes();
    return; 
  }
  var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
  if (!acAddonList) { return; }
  var cat = AMdoc.getElementById("categories");
  var selCat = cat.selectedItem.id;
  selCat = selCat.slice(selCat.indexOf("-") + 1); // remove the 'category-' portion
  addonCatArray = addonsArray[categoryList.indexOf(selCat)];
  
  acAddonList.setAttribute("autocompletesearchparam", JSON.stringify(addonCatArray));
  
  // display ac menu only in categories w/ add-ons of that type
  if (categoryList.indexOf(selCat) != -1) {
    acAddonList.removeAttribute('hidden'); 
  } else {
    acAddonList.setAttribute('hidden', true);
  }
}

function AddonsMgrOnUnload() {
  AddonManager.removeAddonListener(addonListener);
  this.removeEventListener("unload", AddonsMgrOnUnload, false);

  this.removeEventListener("ViewChanged", onViewChange, false);
  var AMdoc = this.document;
  var addonList = AMdoc.getElementById("addon-list");
  addonList.removeEventListener("select", onSelect, false);
  // save selected sort-order; then remove buttons
  var sortButtons = AMdoc.getElementById("aomhilite-sorters");  
  var savedSelections = addonList.getAttribute('aomhilite');
  var sortby = sortButtons.getAttribute("sortby");
  var map = {};
  if (savedSelections) { map = JSON.parse(savedSelections); } 
  map.sortby = sortby;
  var updatedSel = JSON.stringify(map);
  addonList.setAttribute('aomhilite', updatedSel);

  var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                    .getService(ci.nsIPrefService);
  var prefs = prefService.getBranch("extensions.addonsmgrhilte@cfl.");
  aomhiliteMap = addonList.getAttribute('aomhilite');
  prefs.setCharPref("aomhiliteSavedSelections", aomhiliteMap);  
  
  sortButtons.removeEventListener("command", sortButtonHandler, true);
  var listView = AMdoc.getElementById("list-view");
  listView.removeChild(sortButtons);
  
  var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
  if (acAddonList) {
    acAddonList.removeEventListener("textentered", onACaddonListSelection, false);
    var arrow = AMdoc.getAnonymousElementByAttribute(acAddonList, "anonid", "historydropmarker");
    if (arrow && displayDropdown) {
      acAddonList.removeEventListener("click", displayDropdown, false);
      displayDropdown = null;
    }
    var head = AMdoc.getElementById("header");
    head.removeChild(acAddonList);
  }
}

var aomhilitePrefObserver = {
  // based on https://developer.mozilla.org/en-US/Add-ons/Code_snippets/Preferences
  register : function() {
    var prefs = Services.prefs;
    this.branch = prefs.getBranch("extensions.addonsmgrhilte@cfl.");
    this.branch.addObserver("", this, false);
  },

  unregister : function() {
    this.branch.removeObserver("", this);
  },

  observe : function(aSubject, aTopic, aData) {
    // aSubject is the nsIPrefBranch we're observing (after appropriate QI)
    // aData is the name of the pref that's been changed (relative to aSubject)
    switch (aData) {
      case "showVersion":
        showVersionNumber = !showVersionNumber;
        refreshACaddonList();
        break;
      case "showUI":
        showUI = !showUI;
        var mgrs = findAddonsMgrs();
        var i;
        for (i = 0; i < mgrs.length; i++) {
          var AMdoc = mgrs[i].document;
          if (showUI) {
            initUI(AMdoc);
          } else {
            // remove its listbox 
            var header = AMdoc.getElementById("header");          
            var acAddonList = AMdoc.getElementById("aomhilite_ac_menu");
            header.removeChild(acAddonList);
          }
        }
        break;
      case "lastUpdatedStateGrouping":
        lastUpdatedStateGrouping = !lastUpdatedStateGrouping;
        break;        
      case "restore":
      // no action needed since this is only checked once upon AoM load
        break;  
    }
  }
};
